<?php

function flc_cache_add($key, $data, $group = '', $expire = 0) {

    if(wp_using_ext_object_cache()) {
        return wp_cache_add($key, $data, $group, $expire);
    } else {
        return FLC_ObjectCache::instance()->add($key, $data, $group, $expire);
    }
}

function flc_cache_delete($key, $group = '') {
    
    if(wp_using_ext_object_cache()) {
        return wp_cache_delete($key, $group);
    } else {
        return FLC_ObjectCache::instance()->delete($key, $group);
    }
}

function flc_cache_flush() {
    
    if(wp_using_ext_object_cache()) {
        return wp_cache_flush();
    } else {
        return FLC_ObjectCache::instance()->flush();
    }
}

function flc_cache_get($key, $group = '') {
    
    if(wp_using_ext_object_cache()) {
        return wp_cache_get($key, $group);
    } else {
        return FLC_ObjectCache::instance()->get($key, $group);
    }
}

function flc_cache_replace($key, $data, $group = '', $expire = 0) {

    if(wp_using_ext_object_cache()) {
        return wp_cache_replace($key, $data, $group, $expire);
    } else {
        return FLC_ObjectCache::instance()->replace($key, $data, $group, $expire);
    }
}

function flc_cache_set($key, $data, $group = '', $expire = 0) {

    if(wp_using_ext_object_cache()) {
        return wp_cache_set($key, $data, $group, $expire);
    } else {    
        return FLC_ObjectCache::instance()->set($key, $data, $group, $expire);
    }
}

function flc_cache_init() {

    if(wp_using_ext_object_cache()) {
        wp_cache_init();
    }
}

function flc_cache_close() {
    
    if(wp_using_ext_object_cache()) {
        return wp_cache_close();
    } else {
        return FLC_ObjectCache::instance()->save();
    }
}

class FLC_ObjectCache {

    use FLC_Singleton;
    
    var $upload_basedir;
    var $cache_dir;
    var $cache_enabled = false;
    var $expiration_time = 900;
    var $flock_filename = 'flc_object_cache.lock';
    var $mutex;
    var $cache = array();
    var $dirty_objects = array();
    var $non_existant_objects = array();
    var $global_groups = array ('users', 'userlogins', 'usermeta');
    var $non_persistent_groups = array('comment');
    var $blog_id;
    var $cold_cache_hits = 0;
    var $warm_cache_hits = 0;
    var $cache_misses = 0;
    var $secret = '';

    function acquire_lock() {
        // Acquire a write lock.
        $this->mutex = @fopen($this->cache_dir.$this->flock_filename, 'w');
        if ( false == $this->mutex)
            return false;
        flock($this->mutex, LOCK_EX);
        return true;
    }

    function add($key, $data, $group = 'flc', $expire = '') {
        if (empty ($group))
            $group = 'flc';

        if (false !== $this->get($key, $group, false))
            return false;

        return $this->set($key, $data, $group, $expire);
    }

    function delete($key, $group = 'flc', $force = false) {
        if (empty ($group))
            $group = 'flc';

        if (!$force && false === $this->get($key, $group, false))
            return false;

        unset ($this->cache[$group][$key]);
        $this->non_existant_objects[$group][$key] = true;
        $this->dirty_objects[$group][] = $key;
        return true;
    }

    function flush() {
        if ( !$this->cache_enabled )
            return true;

        if ( ! $this->acquire_lock() )
            return false;

        $this->rm_cache_dir();
        $this->cache = array ();
        $this->dirty_objects = array ();
        $this->non_existant_objects = array ();

        $this->release_lock();

        return true;
    }

    function get($key, $group = 'flc', $count_hits = true) {
        if (empty ($group))
            $group = 'flc';

        if (isset ($this->cache[$group][$key])) {
            if ($count_hits)
                $this->warm_cache_hits += 1;
            return $this->cache[$group][$key];
        }

        if (isset ($this->non_existant_objects[$group][$key]))
            return false;

        //  If caching is not enabled, we have to fall back to pulling from the DB.
        if (!$this->cache_enabled) {
            if (!isset ($this->cache[$group]))
                $this->load_group_from_db($group);

            if (isset ($this->cache[$group][$key])) {
                $this->cold_cache_hits += 1;
                return $this->cache[$group][$key];
            }

            $this->non_existant_objects[$group][$key] = true;
            $this->cache_misses += 1;
            return false;
        }

        $cache_file = $this->cache_dir.$this->get_group_dir($group)."/".$this->hash($key);
        if (!file_exists($cache_file)) {
            $this->non_existant_objects[$group][$key] = true;
            $this->cache_misses += 1;
            return false;
        }

        // If the object has expired, remove it from the cache and return false to force
        // a refresh.
        $now = time();
        if ((filemtime($cache_file) + $this->expiration_time) <= $now) {
            $this->cache_misses += 1;
            $this->delete($key, $group, true);
            return false;
        }

        $this->cache[$group][$key] = unserialize(base64_decode(@file_get_contents($cache_file)));
        if (false === $this->cache[$group][$key])
            $this->cache[$group][$key] = '';

        $this->cold_cache_hits += 1;
        return $this->cache[$group][$key];
    }

    function get_group_dir($group) {
        if (false !== array_search($group, $this->global_groups))
            return $group;

        return "{$this->blog_id}/$group";
    }

    function hash($data) {
        if ( function_exists('hash_hmac') ) {
            return hash_hmac('md5', $data, $this->secret);
        } else {
            return md5($data . $this->secret);
        }
    }

    function load_group_from_db($group) {
        return;
    }

    function make_group_dir($group, $perms) {
        $group_dir = $this->get_group_dir($group);
        $make_dir = '';
        foreach (explode('/', $group_dir) as $subdir) {
            $make_dir .= "$subdir/";
            if (!file_exists($this->cache_dir.$make_dir)) {
                if (! @ mkdir($this->cache_dir.$make_dir))
                    break;
                @ chmod($this->cache_dir.$make_dir, $perms);
            }
        }

        return $this->cache_dir."$group_dir/";
    }

    function rm_cache_dir() {
        $dir = $this->cache_dir;
        $dir = rtrim($dir, DIRECTORY_SEPARATOR);
        $top_dir = $dir;
        $stack = array($dir);
        $index = 0;

        while ($index < count($stack)) {
            # Get indexed directory from stack
            $dir = $stack[$index];

            $dh = @ opendir($dir);
            if (!$dh)
                return false;

            while (($file = @ readdir($dh)) !== false) {
                if ($file == '.' or $file == '..')
                    continue;

                if (@ is_dir($dir . DIRECTORY_SEPARATOR . $file))
                    $stack[] = $dir . DIRECTORY_SEPARATOR . $file;
                else if (@ is_file($dir . DIRECTORY_SEPARATOR . $file))
                    @ unlink($dir . DIRECTORY_SEPARATOR . $file);
            }

            $index++;
        }

        $stack = array_reverse($stack);  // Last added dirs are deepest
        foreach($stack as $dir) {
            if ( $dir != $top_dir)
                @ rmdir($dir);
        }

    }

    function release_lock() {
        // Release write lock.
        flock($this->mutex, LOCK_UN);
        fclose($this->mutex);
    }

    function replace($key, $data, $group = 'flc', $expire = '') {
        if (empty ($group))
            $group = 'flc';

        if (false === $this->get($key, $group, false))
            return false;

        return $this->set($key, $data, $group, $expire);
    }

    function set($key, $data, $group = 'flc', $expire = '') {
        if (empty ($group))
            $group = 'flc';

        if (NULL == $data)
            $data = '';

        $this->cache[$group][$key] = $data;
        unset ($this->non_existant_objects[$group][$key]);
        $this->dirty_objects[$group][] = $key;

        return true;
    }

    function save() {
        //$this->stats();

        if (!$this->cache_enabled)
            return true;

        if (empty ($this->dirty_objects))
            return true;

        // Give the new dirs the same perms as upload_basedir.
        $stat = stat($this->upload_basedir);
        $dir_perms = $stat['mode'] & 0007777; // Get the permission bits.
        $file_perms = $dir_perms & 0000666; // Remove execute bits for files.

        // Make the base cache dir.
        if (!file_exists($this->cache_dir)) {
            if (! @ mkdir($this->cache_dir))
                return false;
            @ chmod($this->cache_dir, $dir_perms);
        }

        if ( ! $this->acquire_lock() )
            return false;

        // Loop over dirty objects and save them.
        $errors = 0;
        foreach ($this->dirty_objects as $group => $keys) {
            if ( in_array($group, $this->non_persistent_groups) )
                continue;

            $group_dir = $this->make_group_dir($group, $dir_perms);

            $keys = array_unique($keys);
            foreach ($keys as $key) {
                $cache_file = $group_dir.$this->hash($key);

                // Remove the cache file if the key is not set.
                if (!isset ($this->cache[$group][$key])) {
                    if (file_exists($cache_file))
                        @ unlink($cache_file);
                    continue;
                }

                $temp_file = tempnam($group_dir, 'tmp');
                $serial = base64_encode(serialize($this->cache[$group][$key]));
                $fd = @fopen($temp_file, 'w');
                if ( false === $fd ) {
                    $errors++;
                    continue;
                }
                fputs($fd, $serial);
                fclose($fd);
                if (!@ rename($temp_file, $cache_file)) {
                    if (!@ copy($temp_file, $cache_file))
                        $errors++;
                    @ unlink($temp_file);
                }
                @ chmod($cache_file, $file_perms);
            }
        }

        $this->dirty_objects = array();

        $this->release_lock();

        if ( $errors )
            return false;

        return true;
    }

    function stats() {
        echo "<p>";
        echo "<strong>Cold Cache Hits:</strong> {$this->cold_cache_hits}<br />";
        echo "<strong>Warm Cache Hits:</strong> {$this->warm_cache_hits}<br />";
        echo "<strong>Cache Misses:</strong> {$this->cache_misses}<br />";
        echo "</p>";

        foreach ($this->cache as $group => $cache) {
            echo "<p>";
            echo "<strong>Group:</strong> $group<br />";
            echo "<strong>Cache:</strong>";
            echo "<pre>";
            print_r($cache);
            echo "</pre>";
            if (isset ($this->dirty_objects[$group])) {
                echo "<strong>Dirty Objects:</strong>";
                echo "<pre>";
                print_r(array_unique($this->dirty_objects[$group]));
                echo "</pre>";
                echo "</p>";
            }
        }
    }

    function __construct() {
        global $blog_id;
        
        $upload_dir_arr = wp_upload_dir();
        $this->upload_basedir = $upload_dir_arr['basedir'];

        register_shutdown_function(array(&$this, "__destruct"));

        if (defined('CACHE_PATH'))
            $this->cache_dir = CACHE_PATH;
        else {
            // Using the correct separator eliminates some cache flush errors on Windows
            $this->cache_dir = $this->upload_basedir.DIRECTORY_SEPARATOR.'flc'.DIRECTORY_SEPARATOR.'cache'.DIRECTORY_SEPARATOR;
        }


        if (is_writable($this->cache_dir) && is_dir($this->cache_dir)) {
                $this->cache_enabled = true;
        } else {
            if (is_writable($this->upload_basedir)) {
                $this->cache_enabled = true;
            }
        }

        if (defined('CACHE_EXPIRATION_TIME'))
            $this->expiration_time = CACHE_EXPIRATION_TIME;

        if ( defined('WP_SECRET') )
            $this->secret = WP_SECRET;
        else
            $this->secret = DB_PASSWORD . DB_USER . DB_NAME . DB_HOST . ABSPATH;

        $this->blog_id = $this->hash($blog_id);
    }

    function __destruct() {
        $this->save();
        return true;
    }
}
